<?php


// Turn off all error reporting
error_reporting(0);

/*
 * 2016 - 2017 MIT Janis Dähne
 */

/*
 * version 0.9
 */

#requires sql helper
include_once './sqlHelper.php';

#requires some helper function to e.g. create / delete dirs, handle output...
require './helpers.php';

#require some constants
include_once './constants.php';

include_once './transactionHelper.php';

/**
 * checks if a file name is valid the most basic case is letters, numbers, underscore
 * @param $file a file name
 * @return bool
 */
function is_valid_fileName($file)
{
    return preg_match('/^[a-zA-Z0-9_]+$/', $file) > 0;
}

if ($isDebug) {
    $time_pre = microtime(true);
}

# checks ip and reads the config
$config = require './bootstrap.php';

if ($isDebug) {
    $time_post = microtime(true);
    $exec_time = ($time_post - $time_pre) * 1000; //in ms
    debug("time to read & decode config & check ip in ms: " . $exec_time);
}


if ($isDebug) {
    $time_pre = microtime(true);
}
#request was allowed and config was read

# ============================================================== now check if we have all arguments that we need...

$entityBody = file_get_contents('php://input');

$body = json_decode($entityBody, TRUE);

if ($body === NULL) {
    output($status_code_ArgumentInvalidOrMissing, "could not decode request body");
    exit(1);
}


if (!isset($body[$s_arg_mainFileNameWithExtension])) {
    output($status_code_ArgumentInvalidOrMissing, "argument '" . $s_arg_mainFileNameWithExtension . "' missing");
    exit(1);
}
# the main file name
$arg_mainFileNameWithExtension = $body[$s_arg_mainFileNameWithExtension];


if (!isset($body[$s_arg_allFiles])) {
    output($status_code_ArgumentInvalidOrMissing, "argument '" . $s_arg_allFiles . "' missing");
    exit(1);
}
# all files to use
$arg_allFiles = $body[$s_arg_allFiles];


# check/ensure the iteration dir for the current iteration

$hardGlobalTimeoutInMs = $config[$s_arg_hardTimeoutInMs];

if ($hardGlobalTimeoutInMs === NULL) {
    output($status_code_InternalServerError, 'could not read global hard timeout in ms from config');
    exit(1);
}

$hardMemoryLimitInKb = $config[$s_arg_hardMemoryLimitInKb];

if ($hardMemoryLimitInKb === NULL) {
    output($status_code_InternalServerError, 'could not read global hard memory limit in kb from config');
    exit(1);
}

$hardDiskSpaceLimitInKb = $config[$s_arg_hardDiskSpaceLimitInKb];

if ($hardDiskSpaceLimitInKb === NULL) {
    output($status_code_InternalServerError, 'could not read global hard disk space limit in kb from config');
    exit(1);
}

$requestDistinctionString = NULL;
# try to get the request distinction key/number
if (!isset($body[$s_requestDistinctionNumber])) {
    # client did not set a request distinction number...
    # ignore and hope microtime will handle...
} else {

    if (is_valid_fileName($body[$s_requestDistinctionNumber])) {
        # a valid file name as request id (used to create a dir)
        $requestDistinctionString = $body[$s_requestDistinctionNumber];
    } else {
        output($status_code_InternalServerError, 'request distinct number is not a regex with only letters, numbers and underscores');
        exit(1);
    }
}


if ($isDebug) {
    $time_post = microtime(true);
    $exec_time = ($time_post - $time_pre) * 1000; //in ms
    debug("time to decode body: " . $exec_time);
}

if ($isDebug) {
    $time_pre = microtime(true);
}
$rootDirName = $rootDirNameToUseString;

if ($rootDirName === NULL) {
    goto handleCleanupWithoutDir;
}

if ($isDebug) {
    $time_post = microtime(true);
    $exec_time = ($time_post - $time_pre) * 1000; //in ms
    debug("time to check the next root dir: " . $exec_time);
}

if ($isDebug) {
    $time_pre = microtime(true);
}
# get the transaction
$conn = connectToDb($config['dbServer'], $config['dbName'], $config['dbUser'], $config['dbPw']);

if ($conn === NULL) {
    output($status_code_DbIssue, 'could not connect to db');
    exit(1);
}

$transactionKey = startTransaction($conn, $config, $rootDirName, 'var', $requestDistinctionString);

if ($isDebug) {
    $time_post = microtime(true);
    $exec_time = ($time_post - $time_pre) * 1000; //in ms
    debug("time to open db connection & start the transaction: " . $exec_time);
}

# other arguments depends on the command argument


if (!isset($body[$s_arg_pLang])) {
    endTransaction($conn, $config, $transactionKey, $rootDirName);
    output($status_code_ArgumentInvalidOrMissing, "argument '" . $s_arg_pLang . "' missing");
    exit(1);
}

if (!isset($body[$s_arg_pLang])) {
    endTransaction($conn, $config, $transactionKey, $rootDirName);
    output($status_code_ArgumentInvalidOrMissing, "argument '" . $s_arg_pLang . "' missing");
    exit(1);
}

$arg_pLang = $body[$s_arg_pLang];


# check if the p lang is supported...
# ...get the script to execute from the db like domjudge

$sql = "SELECT * FROM `" . $config['dbTableName'] . "`;";
$allRows = executeSqlAndReturnValue($conn, $sql, []);
$compileCmd = NULL;
$execCmd = NULL;
$sourceFileExtensions = [];

foreach ($allRows as $row) {

    if ($row[$s_plang_internalName] === $arg_pLang) {
        //ok found p lang commands
        $compileCmd = $row[$s_compileCommand];
        $execCmd = $row[$s_executeCommand];

        $sourceFileExtensions = json_decode($row[$s_source_file_extensions], TRUE);

        if ($sourceFileExtensions === NULL) {
            if ($isDebug) {
                fwrite(STDOUT, 'could not decode extensions row');
            }
            endTransaction($conn, $config, $transactionKey, $rootDirName);
            output($status_code_DbIssue, "invalid extensions");
            exit(1);
        }
        break;
    }
}


# create source files
$rootDir = $rootDirName;
$paths = ensureDirs($rootDir, $transactionKey);

if ($paths === FALSE) {
    endTransaction($conn, $config, $transactionKey, $rootDirName);
    output($status_code_FileSystemIssue, "could not create dirs");
    exit(1);
}

$fullWorkingDirPath = $paths[$s_fullWorkingDirPath];

if (createFiles($arg_allFiles, $fullWorkingDirPath) === FALSE) {
    endTransaction($conn, $config, $transactionKey, $rootDirName);
    //error message is in createFiles function
    exit(1);
}

if ($isDebug) {
    debug("all files created!");
}

$testCases = $body['tests'];
$testResults = [];

$testCount = 0;

//check if we already tied to compile
//if the first test is a regex test it wont compile...
$hasTriedToCompile = FALSE;

//store the compile result in case we only copy the result (output for other tests)
$compileResult = array(); //the result of the test
$compileRawResult = array(); //the raw result from do blackbox test

foreach ($testCases as $test) {

    if ($testCount >= $config[$s_arg_maxNumberOfTestsWithOneRequest]) {
        break;
    }

    $testId = $test[$s_test_id];
    $testContent = $test[$s_test_content];
    $testAssets = $test[$s_test_allAssets];


    if (!isset($test[$s_test_command]) || !isset($test[$s_test_id])
        //its ok if we have test content
        //but throw if we have no content && test type is NOT compile (compile do not need test content...)
        || (!isset($test[$s_test_content]) && $test[$s_test_command] !== $s_command_compile)
    ) {

        #output($status_code_ArgumentInvalidOrMissing, "argument '" . $s_test_command . "' missing");
        $missingArgument = $s_test_command;

        if (!isset($test[$s_test_command])) {
            $missingArgument = $s_test_command;
        } else if (!isset($test[$s_test_content])) {
            $missingArgument = $s_test_content;
        } else if (isset($test[$s_test_id])) {
            $missingArgument = $s_test_id;
        }

        $result = array(
            $s_testServerCode => $status_code_ArgumentInvalidOrMissing,
            $s_testServerMessage => "argument '" . $missingArgument . "' missing",
            $s_test_id => $testId,
            $s_passed => FALSE,
            $s_hasCompiled => FALSE, // all other errors returned before this line
            $s_programExitCode => $default_not_executed_user_program_exit_code, //the default value if we not executed the user program
            $s_testResultCode => 100, # this is the result from the test runner program
            $s_protocol => '' # the conversation protocol
        );
        array_push($testResults, $result);
        continue;
    }

    # the command to execute
    $arg_command = $test[$s_test_command];

    //check if valid/supported command
    if (!in_array($arg_command, $s_supportedCommands)) {

        $result = array(
            $s_testServerCode => $status_code_ArgumentInvalidOrMissing,
            $s_testServerMessage => "unknown command, supported commands: [" . implode(', ', $s_supportedCommands) . ']',
            $s_test_id => $testId,
            $s_passed => FALSE,
            $s_hasCompiled => FALSE, // all other errors returned before this line
            $s_programExitCode => $default_not_executed_user_program_exit_code,
            $s_testResultCode => 100, # this is the result from the test runner program
            $s_protocol => '' # the conversation protocol
        );
        array_push($testResults, $result);
        continue;
    }

    if ($arg_command === $s_command_compile) {


        if ($compileCmd === NULL) {
            $result = array(
                $s_testServerCode => $status_code_DbIssue,
                $s_testServerMessage => 'could not find compile command',
                $s_test_id => $testId,
                $s_passed => FALSE,
                $s_hasCompiled => FALSE, // all other errors returned before this line
                $s_programExitCode => $default_not_executed_user_program_exit_code,
                $s_testResultCode => 100, # this is the result from the test runner program
                $s_protocol => '' # the conversation protocol
            );
            array_push($testResults, $result);
            continue;
        } else {

            require_once './do_compileTest_func.php';

            $showTestRunnerDebugOutput = FALSE;
            if (isset($config['maxLinesToRead'])) {
                $showTestRunnerDebugOutput = $config['showTestRunnerDebugOutput'];
            }

            try {
                $_result = do_compile($arg_mainFileNameWithExtension, $fullWorkingDirPath, $compileCmd, $sourceFileExtensions, $requestDistinctionString, $showTestRunnerDebugOutput);
            }catch (Exception $e) {
                //critical error, abort execution (all other tests are likely to fail)
                output($status_code_InternalServerError, "error executing regex test: " . $e->getMessage());
                goto handleCleanup;
            }

            $result = array(
                $s_testServerCode => $status_code_Ok,
                $s_testServerMessage => '',
                $s_test_id => $testId,
                $s_passed => $_result[$s_return_val] === 0,
                $s_hasCompiled => $_result[$s_return_val] !== 50, // all other errors returned before this line
                $s_programExitCode => $default_not_executed_user_program_exit_code,
                $s_testResultCode => $_result[$s_return_val], # this is the result from the test runner program
                $s_protocol => formatOutput($_result[$s_output]) # the conversation protocol
            );
            array_push($testResults, $result);

            //even if the compilation failed we tried so set $hasCompiled to true (others would faild too)
            $hasTriedToCompile = TRUE;

        }
    } else if ($arg_command === $s_command_regexTest) {

        require_once './do_regexTest_func.php';

        $showTestRunnerDebugOutput = FALSE;
        if (isset($config['maxLinesToRead'])) {
            $showTestRunnerDebugOutput = $config['showTestRunnerDebugOutput'];
        }

        try {
            //no need to write any files ... can all be handled in memory
            //$_result = do_regexTest_json($arg_allFiles, $testContent);
            $_result = do_regexTest_format2($arg_allFiles, $testContent, $showTestRunnerDebugOutput);
        } catch (Exception $e) {
            //critical error, abort execution (all other tests are likely to fail)
            output($status_code_InternalServerError, "error executing regex test: " . $e->getMessage());
            goto handleCleanup;
        }


        $result = array(
            $s_testServerCode => $status_code_Ok,
            $s_testServerMessage => '',
            $s_test_id => $testId,
            $s_passed => $_result[$s_return_val] === 0,
            $s_hasCompiled => $_result[$s_return_val] !== 50, // all other errors returned before this line
            $s_programExitCode => $default_not_executed_user_program_exit_code,
            $s_testResultCode => $_result[$s_return_val], # this is the result from the test runner program
            $s_protocol => formatOutput($_result[$s_output]) # the conversation protocol
        );
        array_push($testResults, $result);

    } else if ($arg_command === $s_command_blackBoxTest || $arg_command === $s_command_justRunTest || $arg_command === $s_command_compareFilesTest) {

        if ($isDebug) {
            $time_pre = microtime(true);
        }

        if (!isset($test[$s_test_timeoutInMs])) {
            $result = array(
                $s_testServerCode => $status_code_ArgumentInvalidOrMissing,
                $s_testServerMessage => "argument '" . $s_test_timeoutInMs . "' missing",
                $s_test_id => $testId,
                $s_passed => FALSE,
                $s_hasCompiled => FALSE, // all other errors returned before this line
                $s_programExitCode => $default_not_executed_user_program_exit_code,
                $s_testResultCode => 100, # this is the result from the test runner program
                $s_protocol => '' # the conversation protocol
            );
            array_push($testResults, $result);
            continue;
        }

        # the timeout for the test
        $arg_timeout = $test[$s_test_timeoutInMs];


        if (!isset($test[$s_test_memoryLimitInKb])) {
            $result = array(
                $s_testServerCode => $status_code_ArgumentInvalidOrMissing,
                $s_testServerMessage => "argument '" . $s_test_memoryLimitInKb . "' missing",
                $s_test_id => $testId,
                $s_passed => FALSE,
                $s_hasCompiled => FALSE, // all other errors returned before this line
                $s_programExitCode => $default_not_executed_user_program_exit_code,
                $s_testResultCode => 100, # this is the result from the test runner program
                $s_protocol => '' # the conversation protocol
            );
            array_push($testResults, $result);
            continue;
        }

        # the memory limit for the test
        $arg_MemoryLimit = $test[$s_test_memoryLimitInKb];


        if (!isset($test[$s_test_diskSpaceLimitInKb])) {
            $result = array(
                $s_testServerCode => $status_code_ArgumentInvalidOrMissing,
                $s_testServerMessage => "argument '" . $s_test_diskSpaceLimitInKb . "' missing",
                $s_test_id => $testId,
                $s_passed => FALSE,
                $s_hasCompiled => FALSE, // all other errors returned before this line
                $s_programExitCode => $default_not_executed_user_program_exit_code,
                $s_testResultCode => 100, # this is the result from the test runner program
                $s_protocol => '' # the conversation protocol
            );
            array_push($testResults, $result);
            continue;
        }

        # the max dis space the test can write to
        $arg_DiskSpaceLimit = $test[$s_test_diskSpaceLimitInKb];


        if ($compileCmd === NULL || $execCmd === NULL) {
            $result = array(
                $s_testServerCode => $status_code_DbIssue,
                $s_testServerMessage => "could not find compile command or execute command (unknown plang?)",
                $s_test_id => $testId,
                $s_passed => FALSE,
                $s_hasCompiled => FALSE, // all other errors returned before this line
                $s_programExitCode => $default_not_executed_user_program_exit_code,
                $s_testResultCode => 100, # this is the result from the test runner program
                $s_protocol => '' # the conversation protocol
            );
            array_push($testResults, $result);
            continue;
        }

        # ensure the hard limits

        $min_timeout = $hardGlobalTimeoutInMs;

        if ($arg_timeout) {
            $min_timeout = min($arg_timeout, $hardGlobalTimeoutInMs);
        }

        $min_memoryLimit = $hardMemoryLimitInKb;

        if ($arg_MemoryLimit) {
            $min_memoryLimit = min($arg_MemoryLimit, $hardMemoryLimitInKb);
        }

        $min_diskSpaceLimit = $hardDiskSpaceLimitInKb;

        if ($arg_DiskSpaceLimit) {
            $min_diskSpaceLimit = min($arg_DiskSpaceLimit, $hardDiskSpaceLimitInKb);
        }

        $maxLinesToRead = -1;
        if (isset($config['maxLinesToRead'])) {
            $maxLinesToRead = $config['maxLinesToRead'];
        }

        $maxErrLinesToRead = -1;
        if (isset($config['maxErrLinesToRead'])) {
            $maxErrLinesToRead = $config['maxErrLinesToRead'];
        }

        $maxLinesToWrite = -1;
        if (isset($config['maxLinesToWrite'])) {
            $maxLinesToWrite = $config['maxLinesToWrite'];
        }


        # was just a test...
        #output("t" . $min_timeout . " m" . $min_memoryLimit . " d" . $min_diskSpaceLimit . " dd: " . $arg_DiskSpaceLimit);

        if ($isDebug) {
            $time_post = microtime(true);
            $exec_time = ($time_post - $time_pre) * 1000; //in ms
            debug("time get all args for the black-box-test: " . $exec_time);
        }

        //only the first test needs to compile sources ...

        $needCompilation = $hasTriedToCompile ? FALSE : TRUE;

        $needCompilation = !$hasTriedToCompile;


        if ($needCompilation === FALSE && $compileResult[$s_hasCompiled] === FALSE) {
            //if the compile fail ... all other tests would give use main file not found...
            //so just copy the compile error from the failed compiled test
            $_result = $compileRawResult;
        } else {


            $showTestRunnerDebugOutput = FALSE;
            if (isset($config['maxLinesToRead'])) {
                $showTestRunnerDebugOutput = $config['showTestRunnerDebugOutput'];
            }

            if ($arg_command === $s_command_compareFilesTest) {

                require_once './do_compareFilesTest_func.php';

                try {
                    $_result = do_compareFilesTest($arg_mainFileNameWithExtension, $test, $fullWorkingDirPath,
                        $min_timeout, $min_memoryLimit, $min_diskSpaceLimit,
                        $compileCmd, $execCmd, $sourceFileExtensions, $needCompilation,
                        $maxLinesToRead, $maxErrLinesToRead, $maxLinesToWrite,
                        $requestDistinctionString,
                        $showTestRunnerDebugOutput
                    );
                } catch (Exception $e) {
                    //critical error, abort execution (all other tests are likely to fail)
                    output($status_code_InternalServerError, "error executing compare files test: " . $e->getMessage());
                    goto handleCleanup;
                }
            }
            else {
                require_once './do_blackBoxTest_func.php';

                try {
                    $_result = do_blackBoxTest($arg_mainFileNameWithExtension, $test, $fullWorkingDirPath,
                        $min_timeout, $min_memoryLimit, $min_diskSpaceLimit,
                        $compileCmd, $execCmd, $sourceFileExtensions, $needCompilation,
                        $maxLinesToRead, $maxErrLinesToRead, $maxLinesToWrite,
                        ($arg_command === $s_command_justRunTest),
                        $requestDistinctionString,
                        $showTestRunnerDebugOutput
                    );
                } catch (Exception $e) {
                    //critical error, abort execution (all other tests are likely to fail)
                    output($status_code_InternalServerError, "error executing black box test: " . $e->getMessage());
                    goto handleCleanup;
                }
            }

        }


        $result = array(
            $s_testServerCode => $status_code_Ok,
            $s_testServerMessage => '',
            $s_test_id => $testId,
            $s_passed => $_result[$s_return_val] === 0,
            $s_hasCompiled => $_result[$s_return_val] !== 50, // all other errors returned before this line
            $s_programExitCode => $_result[$s_user_program_exit_code],
            $s_testResultCode => $_result[$s_return_val], # this is the result from the test runner program
            $s_protocol => formatOutput($_result[$s_output]) # the conversation protocol
        );

        if ($needCompilation) {
            //this tried to compile ... store result
            $compileResult = $result;
            $compileRawResult = $_result;
        }

        array_push($testResults, $result);

        //even if the compilation failed we tried so set $hasCompiled to true (others would fail too)
        $hasTriedToCompile = TRUE;
    }

    $testCount++;
}

handleCleanup:
deleteDir($fullWorkingDirPath);
handleCleanupWithoutDir:
# release the transaction
endTransaction($conn, $config, $transactionKey, $rootDirName);

output($status_code_Ok, '', $testResults);

